﻿using System;
using ComponentFramework.Components;
using ComponentFramework.Core;
using ComponentFramework.Structures;
using ComponentFramework.Tools;
using MTV3D65;
using SlimDX;

namespace Stereoscopy
{
    [AutoLoad]
    class FirstPersonCamera : Component, ICameraService
    {
        const float MovementSpeed = 25;
        const float CameraRotationSpeed = 1 / 400.0f;

        TVCamera camera;
        Vector2 cameraAngles;

        public Vector3 Position
        {
            get { return camera.GetPosition().ToVector3(); }
            set { camera.SetPosition(value.X, value.Y, value.Z); }
        }

        public Matrix View
        {
            get { return camera.GetMatrix().ToMatrix(); }
            set { camera.SetMatrix(value.ToTvMatrix()); }
        }

        public float NearPlane
        {
            get { return 0.005f; }
        }
        public float FarPlane
        {
            get { return 1000; }
        }

        float angleOfView;
        public float AngleOfView
        {
            get { return angleOfView; }
            set
            {
                var changed = value != angleOfView;
                angleOfView = value;
                if (changed)
                    UpdateFrustum();
            }
        }

        public float FocalPlane { get; set; }

        public FirstPersonCamera(ICore core) : base(core)
        {
            RevertToDefaults();
        }

        public void RevertToDefaults()
        {
            AngleOfView = 60;
        }

        void UpdateFrustum()
        {
            if (camera == null) return;
            camera.SetViewFrustum(AngleOfView, FarPlane, NearPlane);
        }

        public override void Initialize()
        {
            camera = Core.Engine.GetCamera();
            camera.MoveRelative(-3, 0, 0);
            UpdateFrustum();
        }
        
        public override void Update(TimeSpan elapsedTime)
        {
            bool isRunning = Keyboard.GetKeyState(CONST_TV_KEY.TV_KEY_LEFTSHIFT) == ButtonState.Down;
            float moveSpeed = 0, strafeSpeed = 0;
            float actualSpeed = (isRunning ? 1 : 0.5f) * MovementSpeed * (float)elapsedTime.TotalSeconds;

            if (Keyboard.GetKeyState(CONST_TV_KEY.TV_KEY_W) == ButtonState.Down)
                moveSpeed += actualSpeed;
            if (Keyboard.GetKeyState(CONST_TV_KEY.TV_KEY_S) == ButtonState.Down)
                moveSpeed -= actualSpeed;
            if (Keyboard.GetKeyState(CONST_TV_KEY.TV_KEY_A) == ButtonState.Down)
                strafeSpeed -= actualSpeed;
            if (Keyboard.GetKeyState(CONST_TV_KEY.TV_KEY_D) == ButtonState.Down)
                strafeSpeed += actualSpeed;

            cameraAngles.X = MathHelper.Clamp(cameraAngles.X + Mouse.Movement.Y * CameraRotationSpeed, -MathHelper.HalfPi, MathHelper.HalfPi);
            cameraAngles.Y += Mouse.Movement.X * CameraRotationSpeed;

            camera.SetRotation(cameraAngles.X, cameraAngles.Y, 0);
            camera.MoveRelative(moveSpeed, 0, strafeSpeed);
        }

        public void LookAt(Vector3 point)
        {
            var view = View;

            var forward = Vector3.Normalize(point - view.GetTranslation());
            var right = Vector3.Normalize(Vector3.Cross(view.GetUpVector(), forward));
            var up = Vector3.Normalize(Vector3.Cross(forward, right));

            view.set_Rows(0, new Vector4(right, 0));
            view.set_Rows(1, new Vector4(up, 0));
            view.set_Rows(2, new Vector4(forward, 0));

            View = view;
        }

        [ServiceDependency]
        IKeyboardService Keyboard { get; set; }
        [ServiceDependency]
        IMouseService Mouse { get; set; }
    }
    interface ICameraService : IService
    {
        Vector3 Position { get; set; }
        Matrix View { get; }

        float NearPlane { get; }
        float FarPlane { get; }
        float AngleOfView { get; set; }
        float FocalPlane { get; set; }

        void LookAt(Vector3 position);
        void RevertToDefaults();
    }
}
