﻿#define MULTITHREADING

using System;
using ComponentFramework.Components;
using ComponentFramework.Core;
using ComponentFramework.Tools;
using MTV3D65;
using SlimDX;

namespace Stereoscopy
{
    [AutoLoad]
    class StereoRenderer : Component, IRendererService
    {
        const int MiniDepthSurfaceSize = 128;

        TVRenderSurface miniDepthSurface, leftSurface, rightSurface;
        TVShader combineShader, depthShader;

        TVViewport viewport;
        int viewportWidth, viewportHeight;

        readonly TV_COLOR[] depthPixels = new TV_COLOR[MiniDepthSurfaceSize * MiniDepthSurfaceSize];

        Worker<DownsamplingContext, float> worker;
        readonly DownsamplingContext localContext = new DownsamplingContext { StartRow = 0, EndRow = MiniDepthSurfaceSize / 2 };
        readonly DownsamplingContext workerContext = new DownsamplingContext { StartRow = MiniDepthSurfaceSize / 2, EndRow = MiniDepthSurfaceSize };

        public bool PreRender { get; private set; }

        public float EyeSeparation { get; set; }
        public bool AutoFocus { get; set; }
        public bool ImportanceSampling { get; set; }
        public float FocusAreaFactor { get; set; }

        float redGamma;
        public float RedGamma 
        {
            get { return redGamma; }
            set
            {
                var changed = redGamma != value;
                redGamma = value;
                if (changed)
                    UpdateGamma();
            }
        }

        void UpdateGamma()
        {
            if (combineShader == null) return;
            combineShader.SetEffectParamFloat("RedGamma", RedGamma);
        }

        public StereoRenderer(ICore core) : base(core)
        {
            RevertToDefaults();
        }

        public void RevertToDefaults()
        {
            AutoFocus = true;
            RedGamma = 1.5f;
            EyeSeparation = 0.05f;
            ImportanceSampling = true;
            FocusAreaFactor = 7;
        }

        public override void Initialize()
        {
            viewport = Core.Engine.GetViewport();

            miniDepthSurface = Scene.CreateRenderSurface(MiniDepthSurfaceSize, MiniDepthSurfaceSize, true, CONST_TV_RENDERSURFACEFORMAT.TV_TEXTUREFORMAT_HDR_R32F);
            leftSurface = Scene.CreateRenderSurfaceEx(-1, -1, CONST_TV_RENDERSURFACEFORMAT.TV_TEXTUREFORMAT_A8R8G8B8, true, true, 1);
            rightSurface = Scene.CreateRenderSurfaceEx(-1, -1, CONST_TV_RENDERSURFACEFORMAT.TV_TEXTUREFORMAT_A8R8G8B8, true, true, 1);

            miniDepthSurface.SetSystemMemCopy(true);

            var camera = Core.Engine.GetCamera();
            miniDepthSurface.SetNewCamera(camera);
            leftSurface.SetNewCamera(camera);
            rightSurface.SetNewCamera(camera);

            combineShader = Scene.CreateShader();
            combineShader.CreateFromEffectFile("Combine.fx");
            UpdateGamma();

            depthShader = Scene.CreateShader();
            depthShader.CreateFromEffectFile("Depth.fx");

            worker = ThreadPool.Take<DownsamplingContext, float>(Downsample);
        }

        public override void Update(TimeSpan elapsedTime)
        {
            int newWidth = viewport.GetWidth(), newHeight = viewport.GetHeight();
            if (newWidth != viewportWidth || newHeight != viewportHeight)
            {
                viewportWidth = newWidth;
                viewportHeight = newHeight;
                combineShader.SetEffectParamVector2("TexelSize", new TV_2DVECTOR(1.0f / viewportWidth, 1.0f / viewportHeight));
            }

            // Calculate focal depth for next draw
            if (AutoFocus)
            {
                TextureFactory.GetPixelArrayFloat(miniDepthSurface.GetTexture(), 0, 0, MiniDepthSurfaceSize,
                                                  MiniDepthSurfaceSize, depthPixels);
                worker.Start(workerContext);
                var focalDepthAccumulator = Downsample(localContext);
                Camera.FocalPlane = (worker.Result + focalDepthAccumulator) /
                                    (localContext.ImageWeight + workerContext.ImageWeight);
            }
        }

        class DownsamplingContext
        {
            public int StartRow, EndRow;
            public float ImageWeight;
        }
        float Downsample(DownsamplingContext context)
        {
            var focalDepthAccumulator = 0f;
            context.ImageWeight = 0;

            var farPlane = Camera.FarPlane;
            var invMaxSqDistance = 1 / Math.Pow(MiniDepthSurfaceSize / FocusAreaFactor / 2, 2);
            var surfaceCenter = new Vector2(MiniDepthSurfaceSize, MiniDepthSurfaceSize) / 2;

            for (int row = context.StartRow; row < context.EndRow; row++)
            {
                var rowFirstPixel = row * MiniDepthSurfaceSize;
                for (int col = 0; col < MiniDepthSurfaceSize; col++)
                {
                    var depth = depthPixels[rowFirstPixel + col].r;
                    if (depth == 0)
                        depth = farPlane;

                    float weight;
                    if (ImportanceSampling)
                    {
                        var distanceToCenter = Vector2.DistanceSquared(new Vector2(row, col), surfaceCenter) * invMaxSqDistance;
                        weight = 1 - MathHelper.Saturate(distanceToCenter);
                    }
                    else
                        weight = 1;

                    context.ImageWeight += weight;
                    focalDepthAccumulator += depth * weight;
                }
            }

            return focalDepthAccumulator;
        }

        public override void PreDraw()
        {
            var view = Camera.View;

            var originalPosition = view.GetTranslation();
            var eyeOffset = view.GetRightVector() * EyeSeparation / 2;
            var lookAt = originalPosition + view.GetForwardVector() * Camera.FocalPlane;

            PreRender = true;

            // Left eye
            Camera.Position = originalPosition + eyeOffset;
            Camera.LookAt(lookAt);
            leftSurface.StartRender();
            Core.ReDraw();
            leftSurface.EndRender();

            // Right eye
            Camera.Position = originalPosition - eyeOffset;
            Camera.LookAt(lookAt);
            rightSurface.StartRender();
            Core.ReDraw();
            rightSurface.EndRender();

            // Recenter
            Camera.Position = originalPosition;
            Camera.LookAt(lookAt);

            // Depth pass
            if (AutoFocus)
            {
                depthShader.SetAsGlobal(true);
                {
                    miniDepthSurface.StartRender();
                    Core.ReDraw();
                    miniDepthSurface.EndRender();
                }
                depthShader.SetAsGlobal(false);
            }

            PreRender = false;
        }

        protected override void DisposeInternal()
        {
            ThreadPool.Return(worker);
        }

        public override void PostDraw()
        {
            Screen2DImmediate.Draw_FullscreenQuadWithShader(combineShader, 0, 0, 1, 1, leftSurface.GetTexture(), rightSurface.GetTexture());
        }

        [ServiceDependency]
        ICameraService Camera { get; set; }
        [ServiceDependency]
        IThreadPoolService ThreadPool { get; set; }
    }
    interface IRendererService : IService
    {
        bool PreRender { get; }

        bool AutoFocus { get; set; }
        float RedGamma { get; set; }
        float EyeSeparation { get; set; }
        bool ImportanceSampling { get; set; }
        float FocusAreaFactor { get; set; }

        void RevertToDefaults();
    }
}
